// <<--------------------------------------------->>
// << SLADE (SlayeR's 'LeetAss Doom Editor)       >>
// << By Simon Judd, 2004                         >>
// << ------------------------------------------- >>
// << <keybind.cpp>                               >>
// << KeyBind class functions, handles keys.cfg   >>
// <<--------------------------------------------->>

// Includes ------------------------------------- >>
#include "main.h"
#include "keybind.h"

// BindList::add: Adds a key to the binds list
// ---------------------------------------- >>
void BindList::add(string name, Uint16 key)
{
	keybind_t	*temp = first;
	
	first = new keybind_t;

	first->key = key;
	first->name = name;
	first->next = temp;

	//printf("Added key %s %d\n", name, key);
}

// BindList::key: Gets the key code for a command
// ------------------------------------------- >>
Uint16 BindList::key(string name)
{
	keybind_t	*temp = first;
	
	while (temp)
	{
		if (temp->name == name)
			return temp->key;
		
		temp = temp->next;
	}
	
	return 1;
}

// BindList::pressed: Determines if the key for a command is pressed
// (including any key modifiers shift/ctrl/alt)
// -------------------------------------------------------------- >>
bool BindList::pressed(string name, Uint8* keys)
{
	keybind_t	*temp = first;

	//while (temp->name != name && temp)
		//temp = temp->next;
	while (temp)
	{
		if (temp->name == name)
			break;

		temp = temp->next;
	}

	if (!temp)
		return false;

	if (keys[temp->key])
	{
		bool rightkey = true;

		if (temp->mod & KEYMOD_SHIFT)
		{
			if (!(SDL_GetModState() & KMOD_SHIFT))
				rightkey = false;
		}
		else
		{
			if (SDL_GetModState() & KMOD_SHIFT)
				rightkey = false;
		}

		if (temp->mod & KEYMOD_CTRL)
		{
			if (!(SDL_GetModState() & KMOD_CTRL))
				rightkey = false;
		}
		else
		{
			if (SDL_GetModState() & KMOD_CTRL)
				rightkey = false;
		}

		if (temp->mod & KEYMOD_ALT)
		{
			if (!(SDL_GetModState() & KMOD_ALT))
				rightkey = false;
		}
		else
		{
			if (SDL_GetModState() & KMOD_ALT)
				rightkey = false;
		}

		return rightkey;
	}

	return false;
}

// BindList::load: Loads key binds from a file
// ---------------------------------------- >>
bool BindList::load(char* filename)
{
	//char buffer[128];
	//char temp1[32];
	//char temp2[32];
	MemReader mr;
	string token;
	string name;
	Uint8 mod;

	if (!mr.open_file(filename, 0, 0))
	{
		printf("keys.cfg not found!\n");
		return false;
	}

	token = mr.get_token();

	while (token != "!END")
	{
		mod = 0;
		name = token;

		// Read modifiers
		bool done = false;

		while (!done)
		{
			token = mr.get_token();

			if (token == "SHIFT")
				mod |= KEYMOD_SHIFT;
			else if (token == "CTRL")
				mod |= KEYMOD_CTRL;
			else if (token == "ALT")
				mod |= KEYMOD_ALT;
			else
				done = true;
		}

		// Add key
		Uint16 key = 0;

		if (token == "KEY_UP")
			key = 273;
		else if (token == "KEY_DOWN")
			key = 274;
		else if (token == "KEY_LEFT")
			key = 276;
		else if (token == "KEY_RIGHT")
			key = 275;
		else if (token == "KEY_RETURN")
			key = 13;
		else if (token == "KEY_INSERT")
			key = 277;
		else if (token == "KEY_DELETE")
			key = 127;
		else if (token == "KEY_PGUP")
			key = 280;
		else if (token == "KEY_PGDN")
			key = 281;
		else if (token == "KEY_END")
			key = 279;
		else if (token == "KEY_HOME")
			key = 278;
		else if (token == "KEY_SPACE")
			key = 32;
		else if (token == "KEY_ESCAPE")
			key = 27;
		else if (token == "KEYPAD_UP")
			key = 264;
		else if (token == "KEYPAD_DOWN")
			key = 258;
		else if (token == "KEYPAD_LEFT")
			key = 260;
		else if (token == "KEYPAD_RIGHT")
			key = 262;
		else if (token == "KEYPAD_PLUS")
			key = 270;
		else if (token == "KEYPAD_MINUS")
			key = 269;
		else
			key = (Uint16)token[0];

		add(name, key);
	
		// Apply modifiers
		first->mod = mod;

		string mods;

		if (mod & KEYMOD_SHIFT)
			mods += "SHIFT";
		if (mod & KEYMOD_CTRL)
			mods += "CTRL";
		if (mod & KEYMOD_ALT)
			mods += "ALT";

		//printf("Added key %s = %d, with modifiers %s\n", name.c_str(), first->key, mods.c_str());

		token = mr.get_token();
	}

	return true;
}
