#include "main.h"
#include "map.h"

linegroup_t	lgroups[64];
BYTE		n_lgroups = 0;
char		c_lgroup = 0;

bool		tagged = false;
char		def_argtag = 0;

linetype_t	line_unknown;

extern Map map;
extern stringlist_t line_groups;
extern stringlist_t line_types[64];

linedef_t::linedef_t()
{
	// Common info
	vertex1 = 0;
	vertex2 = 0;
	flags = 0;
	type = 0;
	side1 = -1;
	side2 = -1;
	sector_tag = 0;

	for (int a = 0; a < 5; a++)
		args[a] = 0;

	//hilighted = selected = moving = false;
}

linetype_t::linetype_t()
{
	if (tagged)
		arg_tag = -1;
	else
		arg_tag = def_argtag;

	name = "Unknown Type";
	type = 0;
	args[0] = "Arg1";
	args[1] = "Arg2";
	args[2] = "Arg3";
	args[3] = "Arg4";
	args[4] = "Arg5";
	
	for (int a = 0; a < 5; a++)
		arg_types[a] = NULL;
}

void parse_line(MemReader *mr)
{
	linetype_t *ltype = new linetype_t;

	// Read line type
	ltype->type = mr->get_integer();

	// Read type name
	ltype->name = mr->get_token();

	// Check for extra info
	if (mr->peek_token() == "{")
	{
		string token = mr->get_token();

		while (token != "}")
		{
			if (token == "arg1")
				ltype->args[0] = mr->get_token();

			if (token == "arg2")
				ltype->args[1] = mr->get_token();

			if (token == "arg3")
				ltype->args[2] = mr->get_token();

			if (token == "arg4")
				ltype->args[3] = mr->get_token();

			if (token == "arg5")
				ltype->args[4] = mr->get_token();

			if (token == "arg1_type")
				ltype->arg_types[0] = new string(mr->get_token());

			if (token == "arg2_type")
				ltype->arg_types[1] = new string(mr->get_token());

			if (token == "arg3_type")
				ltype->arg_types[2] = new string(mr->get_token());

			if (token == "arg4_type")
				ltype->arg_types[3] = new string(mr->get_token());

			if (token == "arg5_type")
				ltype->arg_types[4] = new string(mr->get_token());

			if (token == "arg_tag")
				ltype->arg_tag = mr->get_integer();

			token = mr->get_token();
		}
	}

	//ltype->dump();
	lgroups[c_lgroup].add(ltype);
	line_types[c_lgroup].add(ltype->name);
}

int get_lgroup(string name)
{
	int group = -1;

	for (BYTE g = 0; g < n_lgroups; g++)
	{
		if (lgroups[g].name == name)
			group = g;
	}

	return group;
}

void parse_line_group(MemReader *mr)
{
	string	token;

	// Read group name
	token = mr->get_token();
	c_lgroup = get_lgroup(token);

	if (c_lgroup < 0)
	{
		lgroups[n_lgroups].name = token;
		c_lgroup = n_lgroups;
		n_lgroups++;
		line_groups.add(token);
		//printf("Added group \"%s\"\n", lgroups[c_lgroup].name.c_str());
	}
	//else
		//printf("Add to group \"%s\"\n", lgroups[c_lgroup].name.c_str());

	// Read opening brace
	if (mr->get_token() != "{")
	{
		message_box("Error reading line group \"" + lgroups[c_lgroup].name + "\": { expected.");
		return;
	}

	while (token != "}")
	{
		if (token == "tagged")
			tagged = mr->get_bool();

		if (token == "type")
			parse_line(mr);

		if (token == "arg_tag")
			def_argtag = mr->get_integer();

		token = mr->get_token();
	}
}

void line_settype(BYTE group, WORD item, DWORD line)
{
	map.lines[line]->type = lgroups[group].types[item]->type;
}

// Yeah this is probably out of place, but I wasn't thinking ahead when
// I called these 'line types' instead of 'action specials' :P
void thing_setspecial(BYTE group, WORD item, DWORD thing)
{
	map.things[thing]->special = lgroups[group].types[item]->type;
}

linetype_t* get_ltype(BYTE type)
{
	for (DWORD g = 0; g < n_lgroups; g++)
	{
		for (DWORD t = 0; t < lgroups[g].n_types; t++)
		{
			if (lgroups[g].types[t]->type == type)
				return lgroups[g].types[t];
		}
	}

	return &line_unknown;
}