// General structs
#include <string>
using std::string;

// rgba_t: A 32-bit colour definition
struct rgba_t
{
	BYTE r, g, b, a;

	// Constructors
	rgba_t() { r = 0; g = 0; b = 0; a = 0; }
	
	rgba_t(BYTE R, BYTE G, BYTE B, BYTE A)
	{
		r = R;
		g = G;
		b = B;
		a = A;
	}

	// Functions
	void set(BYTE R, BYTE G, BYTE B, BYTE A)
	{
		r = R;
		g = G;
		b = B;
		a = A;
	}

	void set(rgba_t colour)
	{
		r = colour.r;
		g = colour.g;
		b = colour.b;
		a = colour.a;
	}

	float fr() { return (float)r / 255.0f; }
	float fg() { return (float)g / 255.0f; }
	float fb() { return (float)b / 255.0f; }
	float fa() { return (float)a / 255.0f; }
};

// point2_t: A 2d point
struct point2_t
{
	int x, y;

	// Constructors
	point2_t() { x = 0; y = 0; }

	point2_t(int X, int Y)
	{
		x = X;
		y = Y;
	}

	// Functions
	void set(int X, int Y)
	{
		x = X;
		y = Y;
	}

	void set(point2_t p)
	{
		x = p.x;
		y = p.y;
	}

	bool equals(point2_t p)
	{
		if (x == p.x && y == p.y)
			return true;
		else
			return false;
	}
};

// rect_t: A rectangle
struct rect_t
{
	point2_t tl, br;

	// Constructors
	rect_t() { tl.set(0, 0); br.set(0, 0); }

	rect_t(point2_t TL, point2_t BR)
	{
		tl.set(TL);
		br.set(BR);
	}

	rect_t(int x1, int y1, int x2, int y2)
	{
		tl.set(x1, y1);
		br.set(x2, y2);
	}

	rect_t(int x, int y, int width, int height, int align)
	{
		if (align == 0) // Top-left
		{
			tl.set(x, y);
			br.set(x + width, y + height);
		}

		if (align == 1) // Centered
		{
			tl.set(x - (width / 2), y - (height / 2));
			br.set(x + (width / 2), y + (height / 2));
		}
	}

	// Functions
	void set(point2_t TL, point2_t BR)
	{
		tl.set(TL);
		br.set(BR);
	}

	void set(int x1, int y1, int x2, int y2)
	{
		tl.set(x1, y1);
		br.set(x2, y2);
	}

	void set(rect_t rect)
	{
		tl.set(rect.tl);
		br.set(rect.br);
	}

	int x1() { return tl.x; }
	int y1() { return tl.y; }
	int x2() { return br.x; }
	int y2() { return br.y; }

	int width() { return br.x - tl.x; }
	int height() { return br.y - tl.y; }
};

#define RECT_TOPLEFT	0
#define RECT_CENTER		1

struct stringlist_t
{
	string	**strings;
	DWORD	n_strings;
	size_t	max_len;

	stringlist_t() { strings = NULL; n_strings = 0; max_len = 0; }

	void add(string str)
	{
		if (n_strings == 0)
			max_len = str.length();
		else
		{
			if (str.length() > max_len)
				max_len = str.length();
		}

		strings = (string **)realloc(strings, (n_strings + 1) * sizeof(string *));
		strings[n_strings] = new string(str);
		n_strings++;
	}

	void clear()
	{
		for (DWORD s = 0; s < n_strings; s++)
			free(strings[s]);

		free(strings);
		n_strings = 0;
		strings = NULL;
	}

	void sort()
	{
		bool done = false;

		while (!done)
		{
			done = true;

			for (DWORD s = 1; s < n_strings; s++)
			{
				if (*strings[s] < *strings[s - 1])
				{
					string* tmp = strings[s];
					strings[s] = strings[s - 1];
					strings[s - 1] = tmp;
					done = false;
				}
			}
		}
	}
};

struct numlist_t
{
	int		*numbers;
	DWORD	n_numbers;

	numlist_t() { numbers = NULL; n_numbers = 0; }
	~numlist_t() { if (numbers) free(numbers); }

	bool exists(int num)
	{
		for (DWORD n = 0; n < n_numbers; n++)
		{
			if (numbers[n] == num)
				return true;
		}

		return false;
	}

	bool add(int num, bool no_dup)
	{
		if (no_dup && exists(num))
			return false;

		n_numbers++;
		numbers = (int *)realloc(numbers, n_numbers * sizeof(int));
		numbers[n_numbers - 1] = num;

		return true;
	}

	void add_numlist(numlist_t *list, bool no_dup)
	{
		for (DWORD a = 0; a < list->n_numbers; a++)
			add(list->numbers[a], no_dup);
	}

	int find_index(int num)
	{
		for (DWORD n = 0; n < n_numbers; n++)
		{
			if (numbers[n] == num)
				return n;
		}

		return -1;
	}

	int find_index_last(int num)
	{
		for (DWORD n = n_numbers - 1; n >= 0; n--)
		{
			if (numbers[n] == num)
				return n;
		}

		return -1;
	}

	void remove(int num)
	{
		int n = find_index_last(num);

		if (n == -1)
			return;

		for (DWORD i = n; i < n_numbers - 1; i++)
			numbers[i] = numbers[i + 1];

		n_numbers--;
		numbers = (int *)realloc(numbers, n_numbers * sizeof(int));
	}

	void remove_last()
	{
		n_numbers--;
		numbers = (int *)realloc(numbers, n_numbers * sizeof(int));
	}

	void clear()
	{
		n_numbers = 0;
		free(numbers);
		numbers = NULL;
	}

	int get_last()
	{
		if (n_numbers > 0)
			return numbers[n_numbers - 1];
		else
			return 0;
	}
};

#define SORT_ASCENDING 1
#define SORT_DESCENDING 2

struct pointlist_t
{
	point2_t	*points;
	DWORD		n_points;

	pointlist_t() { points = NULL; n_points = 0; }
	~pointlist_t() { if (points) free(points); }

	bool exists(point2_t point)
	{
		for (DWORD n = 0; n < n_points; n++)
		{
			if (points[n].x == point.x &&
				points[n].y == point.y)
				return true;
		}

		return false;
	}

	bool add(point2_t point, bool no_dup)
	{
		bool dup = exists(point);

		if (no_dup && dup)
			return false;

		n_points++;
		points = (point2_t *)realloc(points, n_points * sizeof(point2_t));
		points[n_points - 1].set(point);

		if (dup)
			return false;
		else
			return true;
	}

	int find_index(point2_t point)
	{
		for (DWORD n = 0; n < n_points; n++)
		{
			if (points[n].x == point.x &&
				points[n].y == point.y)
				return n;
		}

		return -1;
	}

	void remove(point2_t point)
	{
		int n = find_index(point);

		if (n == -1)
			return;

		for (DWORD i = n; i < n_points - 1; i++)
			points[i] = points[i + 1];

		n_points--;
		points = (point2_t *)realloc(points, n_points * sizeof(point2_t));
	}

	void clear()
	{
		n_points = 0;
		free(points);
		points = NULL;
	}

	point2_t get_last()
	{
		if (n_points > 0)
			return points[n_points - 1];
		else
			return point2_t(0, 0);
	}
};

struct patch_header_t
{
	short	width;
	short	height;
	short	left;
	short	top;
};

struct col_header_t
{
	BYTE	row_offset;
	BYTE	n_pixels;
};