// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004-05               >>
// << ------------------------------------ >>
// << textures.cpp - Texture loading &     >>
// << handling functions                   >>
// << ------------------------------------ >>

// INCLUDES ------------------------------ >>
#include "main.h"
#include "map.h"

// VARIABLES ----------------------------- >>
TextureList		tex_list;
rgba_t			palette[256];
stringlist_t	pnames;
stringlist_t	sprites;

CVAR(String, ui_skin, "default", CVAR_SAVE);
CVAR(Int, gl_filter, 1, CVAR_SAVE);

// EXTERNAL VARIABLES -------------------- >>
extern WadList wads;

// load_editor_textures: Loads basic editor textures (gui, pointer etc)
// ----------------------------------------------------------------- >>
void load_editor_textures()
{
	string skin = ui_skin;

	tex_list.add_file("_notex", "res/no_tex.png", gl_filter, 0);
	tex_list.add_file("_font", "res/font.png", 3, 0);
	tex_list.add_file("_title", "res/title.png", gl_filter, 0);
	//tex_list.add_file("_thing", "res/thing.png", 2, 0);
	tex_list.add_file("_thing", "res/thing_temp.png", 2, 0);
	tex_list.add_file("_sky", "res/sky.png", gl_filter, 0);
	tex_list.add_file("_unknownsprite", "res/no_thing.png", 2, 0);

	tex_list.add_file("_cursor", "res/" + skin + "/cursor.png", gl_filter, 0);
	tex_list.add_file("_box", "res/" + skin + "/box.png", gl_filter, 0);
	tex_list.add_file("_boxhead", "res/" + skin + "/boxhead.png", gl_filter, 0);
	tex_list.add_file("_menu", "res/" + skin + "/menu.png", gl_filter, 0);
	tex_list.add_file("_xhair", "res/" + skin + "/xhair.png", gl_filter, 0);

	tex_list.add_file("_thing_sound", "res/thing_sound.png", 2, 0);
	tex_list.add_file("_thing_spot", "res/thing_spot.png", 2, 0);
	tex_list.add_file("_thing_light", "res/thing_light.png", 2, 0);
	tex_list.add_file("_thing_fountain", "res/thing_fountain.png", 2, 0);
	tex_list.add_file("_thing_slope", "res/thing_slope.png", 2, 0);
}

// read_palette: Reads the palette from a wad
// --------------------------------------- >>
void read_palette(Wad* wad)
{
	Lump *pal = wad->get_lump("PLAYPAL", 0);
	FILE *fp = fopen(wad->path.c_str(), "rb");

	fseek(fp, pal->Offset(), SEEK_SET);
	for (DWORD c = 0; c < 256; c++)
	{
		fread(&palette[c].r, 1, 1, fp);
		fread(&palette[c].g, 1, 1, fp);
		fread(&palette[c].b, 1, 1, fp);
		palette[c].a = 255;
	}

	palette[247].a = 0;
}

// add_patch_to_tex: Adds patch data to a texture
// ------------------------------------------- >>
void add_patch_to_tex(int x, int y, Wad* wad, Lump *patch, texture_t *tex)
{
	patch_header_t	header;
	long			*columns = NULL;
	BYTE			row = 0;
	BYTE			n_pix = 0;
	BYTE			colour = 0;

	FILE *fp = fopen(wad->path.c_str(), "rb");
	fseek(fp, patch->Offset(), SEEK_SET);

	// Get header & offsets
	//fread(header, sizeof(patch_header_t), 1, fp);
	fread(&header.width, 2, 1, fp);
	fread(&header.height, 2, 1, fp);
	fread(&header.left, 2, 1, fp);
	fread(&header.top, 2, 1, fp);
	columns = (long *)calloc(header.width, sizeof(long));
	fread(columns, sizeof(long), header.width, fp);

	// Read data
	for (int c = 0; c < header.width; c++)
	{
		// Go to start of column
		fseek(fp, patch->Offset(), SEEK_SET);
		fseek(fp, columns[c], SEEK_CUR);

		// Read posts
		while (1)
		{
			// Get row offset
			fread(&row, 1, 1, fp);

			if (row == 255) // End of column?
				break;
			
			// Get no. of pixels
			fread(&n_pix, 1, 1, fp);

			// Read pixels
			fread(&colour, 1, 1, fp); // Skip buffer
			for (BYTE p = 0; p < n_pix; p++)
			{
				fread(&colour, 1, 1, fp);
				tex->add_pixel(x + c, y + row + p, palette[colour]);
			}
			fread(&colour, 1, 1, fp); // Skip buffer & go to next row offset
		}
	}

	fclose(fp);
}

void load_sprite(Wad* wad, string name)
{
	patch_header_t	header;
	long			*columns = NULL;
	BYTE			row = 0;
	BYTE			n_pix = 0;
	BYTE			colour = 0;
	Lump*			patch = NULL;

	patch = wad->get_lump(name, wad->sprites[START]);

	if (!patch)
	{
		printf("Sprite %s not found!\n", name.c_str());
		return;
	}

	FILE *fp = fopen(wad->path.c_str(), "rb");
	fseek(fp, patch->Offset(), SEEK_SET);

	// Get header & offsets
	fread(&header.width, 2, 1, fp);
	fread(&header.height, 2, 1, fp);
	fread(&header.left, 2, 1, fp);
	fread(&header.top, 2, 1, fp);
	columns = (long *)calloc(header.width, sizeof(long));
	fread(columns, sizeof(long), header.width, fp);

	tex_list.add_tex(name, header.width, header.height, false, false);
	tex_list.init_tex_data();
	texture_t *tex = tex_list.get_last();

	// Read data
	for (int c = 0; c < header.width; c++)
	{
		// Go to start of column
		fseek(fp, patch->Offset(), SEEK_SET);
		fseek(fp, columns[c], SEEK_CUR);

		// Read posts
		while (1)
		{
			// Get row offset
			fread(&row, 1, 1, fp);

			if (row == 255) // End of column?
				break;
			
			// Get no. of pixels
			fread(&n_pix, 1, 1, fp);

			// Read pixels
			fread(&colour, 1, 1, fp); // Skip buffer
			for (BYTE p = 0; p < n_pix; p++)
			{
				fread(&colour, 1, 1, fp);
				tex->add_pixel(c, row + p, palette[colour]);
			}
			fread(&colour, 1, 1, fp); // Skip buffer & go to next row offset
		}
	}

	fclose(fp);

	tex_list.create_tex(gl_filter, 0);
}

// load_textures_lump: Loads all textures from a TEXTUREx lump
// -------------------------------------------------------- >>
void load_textures_lump(Wad* wad, Lump *lump)
{
	long n_tex = 0;
	long *offsets = NULL;
	long temp = 0;

	short width = 0;
	short height = 0;
	short patches = 0;

	short xoff = 0;
	short yoff = 0;
	short patch = 0;

	FILE *fp = fopen(wad->path.c_str(), "rb");
	fseek(fp, lump->Offset(), SEEK_SET);

	// Get no. of textures and tex info offsets
	fread(&n_tex, 4, 1, fp);
	offsets = (long *)calloc(n_tex, 4);
	fread(offsets, 4, n_tex, fp);
	
	for (int t = 0; t < n_tex; t++)
	{
		// Go to start of tex definition
		fseek(fp, lump->Offset(), SEEK_SET);
		fseek(fp, offsets[t], SEEK_CUR);

		// Read texture name
		char texname[9] = "";
		fread(texname, 1, 8, fp);
		texname[8] = 0;

		// Skip unused stuff
		fread(&temp, 4, 1, fp);

		// Read width & height
		fread(&width, 2, 1, fp);
		fread(&height, 2, 1, fp);

		// Add texture
		tex_list.add_tex(texname, width, height, true, false);
		tex_list.init_tex_data();
		texture_t *tex = tex_list.get_last();
		//printf("Added texture %s\n", texname);

		// Skip more unused stuff
		fread(&temp, 4, 1, fp);

		// Read no. of patches in texture
		fread(&patches, 2, 1, fp);

		// Add patches
		for (int p = 0; p < patches; p++)
		{
			// Read patch info
			fread(&xoff, 2, 1, fp);
			fread(&yoff, 2, 1, fp);
			fread(&patch, 2, 1, fp);

			// Skip unused
			fread(&temp, 4, 1, fp);

			Wad* iwad = wads.get_iwad();
			Lump *patchlump = NULL;

			// Well, patches _should_ be within P(P)_START and P(P)_END, but some popular wads leave these markers out
			// so I have it like this for compatibility reasons (and doom/2.exe doesn't seem to mind missing markers)
			if (wad->patches[START] == -1)
				patchlump = wad->get_lump(*pnames.strings[patch], 0);
			else
				patchlump = wad->get_lump(*pnames.strings[patch], wad->patches[START]);

			if (patchlump)
				add_patch_to_tex(xoff, yoff, wad, patchlump, tex);
			else
			{
				patchlump = iwad->get_lump(*pnames.strings[patch], iwad->patches[START]);

				if (patchlump)
					add_patch_to_tex(xoff, yoff, iwad, patchlump, tex);
				else
				{
					printf("Patch \"%s\" doesn't exist!\n", pnames.strings[patch]->c_str());
					break;
				}
			}
		}

		tex_list.create_tex(gl_filter, 0);
	}
}

/*
// You'd think loading textures from memory using pointers would be much quicker than
// loading them from the wadfile using fread. Well, the method commented out below isn't :/

// add_patch_to_tex: Adds patch data to a texture
// ------------------------------------------- >>
void add_patch_to_tex(int x, int y, BYTE* pdata, texture_t *tex)
{
	// Get header & offsets
	patch_header_t *header = (patch_header_t *)pdata;
	long *col_offsets= (long *)((BYTE *)pdata + sizeof(patch_header_t));

	// Read data
	for (int c = 0; c < header->width; c++)
	{
		// Go to start of column
		BYTE* data = (BYTE *)pdata;
		data += col_offsets[c];

		// Read posts
		while (1)
		{
			// Get row offset
			BYTE row = *data;

			if (row == 255) // End of column?
				break;
			
			// Get no. of pixels
			data++;
			BYTE n_pix = *data;

			data++; // Skip buffer
			for (BYTE p = 0; p < n_pix; p++)
			{
				data++;
				tex->add_pixel(x + c, y + row + p, palette[*data]);
			}
			data += 2; // Skip buffer & go to next row offset
		}
	}
}

// load_textures_lump: Loads all textures from a TEXTUREx lump
// -------------------------------------------------------- >>
void load_textures_lump(Wad* wad, BYTE* lump)
{
	// Get no. of textures and tex info offsets
	long *n_tex = (long *)lump;
	long *offsets = (long *)((BYTE *)lump + 4);

	for (int t = 0; t < *n_tex; t++)
	{
		// Go to start of tex definition
		BYTE *data = (BYTE *)lump;
		data += offsets[t];

		// Read texture name
		char texname[9] = "";
		memcpy(texname, data, 8);
		texname[8] = 0;
		data += 8;

		// Skip unused stuff
		data += 4;

		// Read width & height
		short width = (short)*data;
		data += 2;
		short height = (short)*data;
		data += 2;

		// Add texture
		tex_list.add_tex(texname, width, height, true, false);
		tex_list.init_tex_data();
		texture_t *tex = tex_list.get_last();
		//printf("Added texture %s\n", texname);

		// Skip more unused stuff
		data += 4;

		// Read no. of patches in texture
		short patches = (short)*data;
		data += 2;

		// Add patches
		for (int p = 0; p < patches; p++)
		{
			// Read patch info
			short xoff = (short)*data;
			data += 2;
			short yoff = (short)*data;
			data += 2;
			short patch = (short)*data;
			data += 2;

			// Skip unused
			data += 4;

			Wad* iwad = wads.get_iwad();

			if (wad->get_lump(*pnames.strings[patch], wad->patches[START]))
				add_patch_to_tex(xoff, yoff, wad->get_lump(*pnames.strings[patch], wad->patches[START])->Data(), tex);
				//add_patch_to_tex(xoff, yoff, wad->lump_mem(wad->get_lump_index(*pnames.strings[patch])), tex);
			else
			{
				if (iwad->get_lump(*pnames.strings[patch], iwad->patches[START]))
					add_patch_to_tex(xoff, yoff, iwad->get_lump(*pnames.strings[patch], wad->patches[START])->Data(), tex);
					//add_patch_to_tex(xoff, yoff, iwad->lump_mem(iwad->get_lump_index(*pnames.strings[patch])), tex);
				else
				{
					printf("Patch \"%s\" doesn't exist in IWAD!\n", pnames.strings[patch]->c_str());
					break;
				}
			}
		}

		tex_list.create_tex(gl_filter, 0);
	}
}
*/

// load_pnames: Loads the pnames lump from a wad
// ------------------------------------------ >>
void load_pnames(Wad* wad)
{
	DWORD n_pnames = 0;
	FILE *fp = fopen(wad->path.c_str(), "rb");
	Lump *lump = wad->get_lump("PNAMES", 0);

	fseek(fp, lump->Offset(), SEEK_SET);
	fread(&n_pnames, 4, 1, fp);

	for (DWORD p = 0; p < n_pnames; p++)
	{
		char name[9] = "";

		fread(name, 1, 8, fp);
		name[8] = 0;

		pnames.add(name);
	}

	fclose(fp);
}

// load_textures: Loads all textures from the last wad opened containing a TEXTUREx lump
// (or the IWAD if no open wads have TEXTUREx lumps...)
// ---------------------------------------------------------------------------------- >>
void load_textures()
{
	Wad* tex_wad = wads.get_wad_with_lump("TEXTURE1");

	printf("Loading textures from %s\n", tex_wad->path.c_str());

	// Load PNAMES
	load_pnames(tex_wad);

	// Load TEXTURE1
	//load_textures_lump(tex_wad, tex_wad->get_lump("TEXTURE1", 0)->Data());
	load_textures_lump(tex_wad, tex_wad->get_lump("TEXTURE1", 0));

	// Load TEXTURE2 if it exists
	if (tex_wad->get_lump("TEXTURE2", 0))
		load_textures_lump(tex_wad, tex_wad->get_lump("TEXTURE2", 0));
		//load_textures_lump(tex_wad, tex_wad->get_lump("TEXTURE2", 0)->Data());
}

// load_flats_wad: Loads all flats from a wad
// --------------------------------------- >>
void load_flats_wad(Wad* wad)
{
	texture_t *tex = NULL;
	int lump = wad->flats[START];
	FILE *fp = fopen(wad->path.c_str(), "rb");

	lump++;

	print(true, "%d\n", wad->flats[END]);

	while (lump < wad->flats[END])
	{
		if (wad->directory[lump]->Size() > 0)
		{
			BYTE p = 0;

			// Create the texture
			tex_list.add_tex(wad->directory[lump]->Name(), 64, 64, true, true);
			tex_list.init_tex_data();
			tex = tex_list.get_last();

			// Read flat data
			fseek(fp, wad->directory[lump]->Offset(), SEEK_SET);

			for (BYTE y = 0; y < 64; y++)
			{
				for (BYTE x = 0; x < 64; x++)
				{
					fread(&p, 1, 1, fp);
					tex->add_pixel(x, y, palette[p]);
				}
			}

			// Generate the texture
			tex_list.create_tex(gl_filter, 0);
		}

		lump++;
	}
}

// load_flats: Loads flats from all open wads (that have any flats in them)
// --------------------------------------------------------------------- >>
void load_flats()
{
	print(true, "Load flats...\n");

	// Load IWAD flats first
	load_flats_wad(wads.get_iwad());

	// Load flats from all open pwads
	if (wads.n_wads == 0)
		return;

	for (DWORD w = 0; w < wads.n_wads; w++)
	{
		if (wads.get_wad(w)->flats[START] != -1)
			load_flats_wad(wads.get_wad(w));
	}

	print(true, "ok\n");
}

// load_sprites: Loads sprites for thing previews
// ------------------------------------------- >>
void load_sprites()
{
	for (int s = 0; s < sprites.n_strings; s++)
	{
		Wad* wad = wads.get_wad_with_lump(*sprites.strings[s]);

		if (wad)
		{
			//printf("Loading sprite %s\n", sprites.strings[s]->c_str());
			load_sprite(wad, *sprites.strings[s]);
		}
	}
}

void load_tx_texture(Wad* wad, long index)
{
	// Get lump
	Lump* lump = wad->directory[index];
	BYTE* data = lump->Data();

	// Check for PNG header
	if (data[0] == 137 && data[1] == 80 &&
		data[2] == 78 && data[3] == 71 &&
		data[4] == 13 && data[5] == 10 &&
		data[6] == 26 && data[7] == 10)
	{
		// Write lump content to file
		FILE *fp = fopen("sladetemp", "wb");
		fwrite(lump->Data(), lump->Size(), 1, fp);
		fclose(fp);

		// Read file to texture
		tex_list.add_file(lump->Name(), "sladetemp", gl_filter, 0);
		tex_list.get_tex(lump->Name(), false, false)->available = true;

		remove("sladetemp");
	}
	else // Normal doom format patch
	{
		short *p = (short *)lump->Data();
		short width = *p++;
		short height = *p;

		if (width < 0 || height < 0)
			return;

		tex_list.add_tex(lump->Name(), width, height, true, false);
		tex_list.init_tex_data();
		add_patch_to_tex(0, 0, wad, lump, tex_list.get_last());
		tex_list.create_tex(gl_filter, 0);
	}
}

void load_tx_textures()
{
	for (DWORD w = 0; w < wads.n_wads; w++)
	{
		if (wads.get_wad(w)->tx[START] != -1)
		{
			Wad* wad = wads.get_wad(w);
			bool done = false;
			long lump = wad->tx[START] + 1;

			while (!done)
			{
				load_tx_texture(wad, lump);
				lump++;

				if (lump == wad->tx[END] || lump == wad->num_lumps)
					done = true;
			}
		}
	}
}
